/* Copyright (C) 1995,1996 Robert de Bath <rdebath@cix.compulink.co.uk>
 * This file is part of the Linux-8086 C library and is distributed
 * under the GNU Library General Public License.
 */

/*
 * This deals with both the atexit and on_exit function calls
 * 
 * Note calls installed with atexit are called with the same args as on_exit
 * fuctions; the void* is given the NULL value.
 * 
 * Also if __cleanup already has a value (Not __on_exit) this is put into the
 * table as the next 'atexit' call.
 */

#include <errno.h>

/* ATEXIT.H */
#define MAXONEXIT 20		/* AIUI Posix requires 10 */

typedef void (*vfuncp) ();

extern vfuncp __cleanup;
extern void __do_exit();

extern struct exit_table
{
   vfuncp called;
   void *argument;
}
__on_exit_table[MAXONEXIT];

extern int __on_exit_count;

/* End ATEXIT.H */

#ifdef L_atexit
int
atexit(ptr)
vfuncp ptr;
{
   if (__cleanup && __cleanup != __do_exit)
   {
      vfuncp ptr = __cleanup;
      __cleanup = 0;
      atexit(__cleanup);
   }
   __cleanup = __do_exit;
   if (__on_exit_count >= MAXONEXIT)
   {
      errno = ENOMEM;
      return -1;
   }
   __on_exit_table[__on_exit_count].called = ptr;
   __on_exit_table[__on_exit_count].argument = 0;
   __on_exit_count++;
   return 0;
}

#endif

#ifdef L_on_exit
int
on_exit(ptr, arg)
vfuncp ptr;
void *arg;
{
   if (__cleanup && __cleanup != __do_exit)
   {
      vfuncp ptr = __cleanup;
      __cleanup = 0;
      on_exit(__cleanup, 0);
   }
   __cleanup = __do_exit;
   if (__on_exit_count >= MAXONEXIT)
   {
      errno = ENOMEM;
      return -1;
   }
   __on_exit_table[__on_exit_count].called = ptr;
   __on_exit_table[__on_exit_count].argument = arg;
   __on_exit_count++;
   return 0;
}

#endif

#ifdef L___do_exit

int   __on_exit_count = 0;

void
__do_exit(rv)
int   rv;
{
   int   count = __on_exit_count;
   int   i;
   __on_exit_count = MAXONEXIT;		/* ensure no more will be added */

   /* In reverse order */
   for (i = count - 1; i >= 0; i--)
      (*__on_exit_table[i].called) (rv, __on_exit_table[i].argument);
}

#endif
