/*   fprintf.c for limited Linux libc
     Copyright (C) 1996  Joel N. Weber II <nemo@koa.iolani.honolulu.hi.us>
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/* Thanks Alan for writing the hard routine for me :-)
 * Alan said that this works "most of the time".  Something tells me I'm making
 * it even worse! */
/* The basic idea here is to make fprintf the core routine.  printf obviously
   can just call fprintf with stdout followed by all of its arguments.
   sprintf() works using the fake file &the_sprintf.  It's marked as fully
   buffered, so that it will only write(2) when &the_sprintf->bufpos ==
   &the_sprintf->bufend, which I doubt will happen since &the_sprintf->bufend
   = 0.  The trick is that sprintf must set &the_sprintf->bufstart =
   &the_sprintf->bufpos = its first argument.  Not as orthagonal (is that
   the right word?) as glibc's facilities for non-files, but this isn't a
   library for people who have unlimited RAM budgets.  (not like the libc
   I use on linux/i586 enjoys an unlimited RAM budget either; I only have
   8 MB

   I'm not sure what the "correct" way to pass the variable arguments
   from one function to the next is.  Rather than pass the arguments
   themselves, I'm passing a pointer to them.  However, the following
   explaination from Alan is probably a polite way of saying it will not
   work on a 386 anyway.
   Joel Weber */
/*
 *	This is NOT stunningly portable, but works
 *	for pretty dumb non ANSI compilers and is
 *	tight.	Adjust the sizes to taste.
 *
 *	Illegal format strings will break it. Only the
 *	following simple subset is supported
 *
 *	%x	-	hex
 *	%d	-	decimal
 *	%s	-	string
 *	%c	-	char
 *
 *	And the h/l length specifiers for %d/%x
 *
 *		Alan Cox.
 */

/* One of these days the headers will work painlessly... */
/* #include <linuxmt/types.h> */
/* #include <linuxmt/fcntl.h> */
#include "stdio.h"

/* 17 to make sure that there's room for the trailing newline.
   I'm not really sure if this is ideal... */
static char nstring[17]="0123456789ABCDEF";

static char *__numout(i, base)
long i;
int base;
{
  static unsigned char out[16];
  int n;
  int flg = 0;
  unsigned long val;

  if( i<0 && base==10 ) { flg = 1; i = -i; }
  val = i;

  for (n = 0; n < 15; n++) out[n] = ' ';
  out[15] = '\0';
  n = 14;
  do{
    out[n] = nstring[val % base];
    n--;
    val /= base;
  }
  while(val);
  if(flg) out[n--] = '-';
  return &out[n+1];
}

/* currently, the h/l specifications are ignored */
static int internal_fprintf(stream,fmt,a1)
FILE *stream;
char *fmt;
int *a1;
{
	unsigned char *p=(char *)a1;
	char c;
	int count = 0;
	int type, base;

	while((c=*fmt++))
	{
		if(c!='%'){
                   	putc(c, stream);
			count++;
		}
		else
		{
			type=1;
			c=*fmt++;
			if(c=='h')
			{
				c=*fmt++;
				type = 0;
			}
			else if(c=='l')
			{
				c=*fmt++;
				type = 2;
			}
			
			switch(c)
			{
				case 'x':
					base=16;
					if(0) {
				case 'o': 	base=8;
					}
					if(0) {
				case 'd': 	base=10;
					}
					{
						long val;
						switch(type)
						{
						case 0:
							val= *(short*)p;
							p+=sizeof(short);
							break;
						case 1:
							val= *(int*)p;
							p+=sizeof(int);
							break;
						case 2:
							val= *(long*)p;
							p+=sizeof(long);
							break;
						}
						fputs(__numout(val,base), stream);
					}
					break;
				case 's':
				{
					char *cp=*((char **)p);
					p+=sizeof(char *);
					while(*cp)
						putc(*cp++, stream);
					break;
				}
				case 'c':
					putc(*(int*)p, stream);
					p+=sizeof(int);
					break;
				default:
					putc(c, stream);
			}
		}
	}
	return count;
}


int fprintf(stream,fmt,a1)
FILE *stream;
char *fmt;
int a1;
{
  return internal_fprintf(stream,fmt,&a1);
}

int printf(fmt,a1)
char *fmt;
int a1;
{
  return internal_fprintf(stdout,fmt,&a1);
}

int sprintf(s,fmt,a1)
char *s;
char *fmt;
int a1;
{
/* This is a strange way of doing sprintf, but it should work */
static FILE the_sprintf = {
   -1,
   0,
   0,
   0,
   0,
   _IOFBF,
   _MODE_WRITE,
   0, 0,
   0, 0
};
  int rv;

  the_sprintf.bufstart = s;
  the_sprintf.bufpos = s;
  the_sprintf.fc_err = 0;
  rv = internal_fprintf(&the_sprintf,fmt,&a1);
  putc('\0', &the_sprintf); /* Don't forget the null */
  return rv;
}
